#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bmsg.h"
#include "fullempty_client.h"

/* Implement the "client" side of fullempty variables (in terms of bmsg). */

/* Each fullempty variable is represented by an index into the
   `fullempty_vars' array. */

typedef struct {
  /* Linked list of pending "puts", waiting until the variable is empty: */
  struct put_t *waiting_puts;
  /* Value for a full variable, NULL for an empty: */
  struct put_t *ready_put;
  /* Linked list of pending "gets", waiting until the variable is full: */
  struct get_t *waiting_gets;
} fullempty_var_t;

/* Array sized on demand: */
static fullempty_var_t *fullempty_vars;
static int fullempty_count;

typedef struct put_t {
  int given_by; /* identifies process */
  int data_len;
  struct put_t *next;
  char data[1];
} put_t;

typedef struct get_t {
  int requested_by; /* identifies process */
  int requested_len;
  struct get_t *next;
} get_t;

static void prepare_waiting_puts_waiting_gets(int id)
{
  /* Resizes fullempty_var as needed */

  if (id < 0) {
    fprintf(stderr, "bad emptyfree id %d\n", id);
    abort();
  }
  
  if (fullempty_count <= id) {
    int c;
    fullempty_var_t *fv;
    
    c = fullempty_count * 2;
    if (c <= id)
      c = (id + 1) * 2;
    
    fv = (fullempty_var_t *)calloc(sizeof(fullempty_var_t), c);
    
    memcpy(fv, fullempty_vars, fullempty_count * sizeof(fullempty_var_t));

    if (fullempty_count) {
      free(fullempty_vars);
    }

    fullempty_vars = fv;
    fullempty_count = c;
  }
}

static int any_missing(int s, int t)
{
  int i;

  if (fullempty_count < t)
    return 1;

  for (i = s; i < t; i++)
    if (!fullempty_vars[i].ready_put)
      return 1;

  return 0;
}

void fullempty_loop(int t, struct connection_t **c, int start_check, int end_check)
{
  fullempty_var_t *var;
  int i, a[3], id;

  /* handle put and get actions on fullempty vars: */
  while (any_missing(start_check, end_check)) {
    i = wait_connection_recv(t, c);

    /* receive a put or get: */
    recv_byte_message(c[i], 3 * sizeof(int), a);

    /* Extract id and map it to a fullempty variable record: */
    id = a[1];
    prepare_waiting_puts_waiting_gets(id);
    var = &(fullempty_vars[id]);

    if (a[0] == 1) {
      /* create a pending put */
      put_t *p;
      p = (put_t *)malloc(sizeof(put_t) + a[2]);
      p->given_by = i;
      p->data_len = a[2];
      recv_byte_message(c[i], p->data_len, &p->data);
      p->next = var->waiting_puts;
      var->waiting_puts = p;
    } else {
      /* create a pending get */
      get_t *g;
      g = (get_t *)malloc(sizeof(get_t) + a[2]);
      g->requested_by = i;
      g->requested_len = a[2];
      g->next = var->waiting_gets;
      var->waiting_gets = g;
    }

    if (var->waiting_puts && !var->ready_put) {
      /* var is empty, so accept a pending put: */
      var->ready_put = var->waiting_puts;
      var->waiting_puts = var->waiting_puts->next;
      send_byte_message(c[var->ready_put->given_by], sizeof(int), a); /* confirm receipt */
    }
    if (var->ready_put && var->waiting_gets) {
      /* var is full, so send data for a pending get */
      get_t *next;
      if (var->waiting_gets->requested_len != var->ready_put->data_len) {
        fprintf(stderr,
                "mismatch: get on %d wants %d bytes, but value is %d bytes\n",
                id, var->waiting_gets->requested_len, var->ready_put->data_len);
        abort();
      }
      send_byte_message(c[var->waiting_gets->requested_by], 
                        var->waiting_gets->requested_len, 
                        &var->ready_put->data);
      free(var->ready_put);
      var->ready_put = NULL;
      next = var->waiting_gets->next;
      free(var->waiting_gets);
      var->waiting_gets = next;
    }
  }
}

void fullempty_get(int id, int len, void *dest)
{
  fullempty_var_t *var;

  if (fullempty_count <= id) {
    fprintf(stderr, "no value available for fullempty %d\n", id);
    abort();
  }

  var = &(fullempty_vars[id]);
  if (var->ready_put->data_len != len) {
    fprintf(stderr,
            "fullempty %d has %d bytes, but expected %d bytes\n",
            id, var->ready_put->data_len, len);
    abort();
  }

  memcpy(dest, &var->ready_put->data, len);

  free(var->ready_put);
  var->ready_put = NULL;
}

int fullempty_bytes_available(int id)
{
  fullempty_var_t *var;

  if (fullempty_count <= id)
    return -1;
  
  var = &(fullempty_vars[id]);
  if (var->ready_put)
    return var->ready_put->data_len;
  else
    return -1;
}
